#!/usr/bin/env python
import numpy as np
import paths as pt
from tools import file_writer, data_loader
from sklearn.model_selection import StratifiedKFold
import xgboost as xgb
from sklearn.metrics import accuracy_score
import xgboost as xgb
import pandas as pd
from utility.metrics import gini_xgb
import shap
from typing import List

CASE = "Complete"
COMPLETE_FILENAME = "complete_emb.csv"
COMPLIANCE_FILENAME = "compliance_emb.csv"
FALL_FILENAME = "fall_emb.csv"
CSV_FILENAME = f"{CASE} best features.csv"
PLOT_FILENAME = f"{CASE} SHAP feature values"
NUM_ITERATIONS = 5

def main():
    if CASE == "Complete":
        dl = data_loader.CompleteDataLoader(COMPLETE_FILENAME).load_data()
    else:
        dl = data_loader.FallDataLoader(FALL_FILENAME).load_data()
    
    X, y = dl.get_data()
    cols = X.columns
    X = np.array(X)
    y = np.array(y)

    total_shap_df  = pd.DataFrame()
    for seed in range(NUM_ITERATIONS):
        print('#'*40, '{} of {} iterations'.format(seed+1, NUM_ITERATIONS), '#' * 40)
        temp_shap_df = get_best_shap_features(X, y, cols, seed)
        total_shap_df = pd.concat([total_shap_df, temp_shap_df])
    
    shap_sorted_df = total_shap_df.groupby('feature').mean().sort_values('shap_values',
                                                                         ascending=False)
    shap_sorted_df = shap_sorted_df.reset_index()

    importances = shap_sorted_df['shap_values']
    features = shap_sorted_df['feature']
    file_writer.write_shap_importance_plot(features, importances, pt.REPORTS_PLOTS_DIR, PLOT_FILENAME)
    file_writer.write_csv(shap_sorted_df, pt.REPORTS_DIR, CSV_FILENAME)

def get_best_shap_features(X: np.ndarray, y: np.ndarray,
                           cols: List[str], seed: int):
    K = 5
    kf = StratifiedKFold(n_splits=K, random_state=seed, shuffle=True)
    feat_importance_df  = pd.DataFrame()
    
    neg, pos = np.bincount(y)
    scale_pos_weight = neg / pos

    model = xgb.XGBClassifier(n_estimators=200,
                              learning_rate=0.07,
                              objective='binary:logistic',
                              scale_pos_weight=scale_pos_weight,
                              eval_metric='logloss',
                              use_label_encoder=False,
                              n_jobs=-1,
                              random_state=0,
                              seed=seed)
    
    acc_score_list = list()
    for i, (train_index, test_index) in enumerate(kf.split(X, y)):
        print('='*30, '{} of {} folds'.format(i+1, K), '='*30)
        X_train, X_val = X[train_index], X[test_index]
        y_train, y_val = y[train_index], y[test_index]
        model.fit(X_train, y_train,
                  eval_set=[(X_train, y_train), (X_val, y_val)],
                  eval_metric=gini_xgb,
                  verbose=0)
        y_val_pred = model.predict_proba(X_val)[:,1]
        y_scores_new = (y_val_pred > 0.5)
        acc_score_list.append(np.around(accuracy_score(y_val, y_scores_new), decimals=3))
        shap_values = shap.TreeExplainer(model).shap_values(X_train)
        fold_importance_df  = pd.DataFrame()
        fold_importance_df['feature'] = cols
        fold_importance_df['shap_values'] = np.around(abs(np.array(shap_values)[:,:]).mean(0), decimals=2)
        fold_importance_df['feat_imp'] = np.around(model.feature_importances_, decimals=2)
        feat_importance_df = pd.concat([feat_importance_df, fold_importance_df])
    
    mean_acc_score = np.mean(acc_score_list)
    print(f"Mean accuracy: {mean_acc_score}")
    feat_importance_df_shap = feat_importance_df.groupby('feature').mean().sort_values('shap_values',
                                                                                       ascending=False)
    feat_importance_df_shap = feat_importance_df_shap.reset_index()    
    return feat_importance_df_shap

if __name__ == '__main__':
    main()
