#!/usr/bin/env python
from tools import data_loader, preprocessor
import tensorflow as tf
import numpy as np
from sklearn.model_selection import train_test_split
import matplotlib.pyplot as plt

CASE = "Complete"
COMPLETE_FILENAME = "complete_emb.csv"
COMPLIANCE_FILENAME = "compliance_emb.csv"
FALL_FILENAME = "fall_emb.csv"
COLORS = plt.rcParams['axes.prop_cycle'].by_key()['color']

METRICS = [
      tf.keras.metrics.BinaryAccuracy(name='accuracy'),
      tf.keras.metrics.Precision(name='precision'),
      tf.keras.metrics.Recall(name='recall'),
      tf.keras.metrics.AUC(name='auc'),
]

def make_model(input_dim, output_bias=None):
    if output_bias is not None:
        output_bias = tf.keras.initializers.Constant(output_bias)
    model = tf.keras.models.Sequential()
    model.add(tf.keras.layers.Dense(128,
     input_dim=input_dim,
     activation='relu'))
    model.add(tf.keras.layers.Dropout(0.2))
    model.add(tf.keras.layers.Dense(1, activation='sigmoid',
                                    bias_initializer=output_bias))
    model.compile(loss='binary_crossentropy',
     optimizer=tf.keras.optimizers.Adam(),
     metrics=METRICS)
    return model

def make_plots(history, label, n, metric):
    plt.plot(history.epoch, history.history[metric],
               color=COLORS[n], label='Train ' + label)
    plt.plot(history.epoch, history.history[f'val_{metric}'],
               color=COLORS[n], label='Val ' + label,
               linestyle="--")
    plt.xlabel('Epoch')
    plt.ylabel(metric)
    plt.legend()
  
def main():
    if CASE == "Complete":
        X, y = data_loader.CompleteDataLoader(COMPLETE_FILENAME) \
        .load_data().prepare_data(normalization_strategy="BoxCox",
                                  scaling_strategy="Standard")
    elif CASE == "Compliance":
        X, y = data_loader.ComplianceDataLoader(COMPLIANCE_FILENAME) \
        .load_data().prepare_data()
    else:
        X, y = data_loader.FallDataLoader(FALL_FILENAME) \
        .load_data().prepare_data()
        
    X_train, X_valid, y_train, y_valid = train_test_split(X, y,
     test_size=0.3, random_state=0)
    
    model = make_model(input_dim=X.shape[1])
    no_class_weight_history = model.fit(np.array(X_train), np.array(y_train), epochs=50,
     validation_data=(np.array(X_valid), np.array(y_valid)), batch_size=16, verbose=False)
    
    neg, pos = np.bincount(y_train)
    initial_bias = np.log([pos/neg])
    model = make_model(input_dim=X.shape[1], output_bias=initial_bias)
    no_class_weight_bias_history = model.fit(np.array(X_train), np.array(y_train), epochs=50,
     validation_data=(np.array(X_valid), np.array(y_valid)), batch_size=16, verbose=False)
    
    class_weight = preprocessor.get_class_weight(neg, pos)
    model = make_model(input_dim=X.shape[1])
    class_weight_history = model.fit(np.array(X_train), np.array(y_train), epochs=50,
     validation_data=(np.array(X_valid), np.array(y_valid)), batch_size=16,
     class_weight=class_weight, verbose=False)
    
    make_plots(no_class_weight_history, f"{CASE} no class weight", 0, 'loss')
    make_plots(no_class_weight_bias_history, f"{CASE} no class weight bias", 1, 'loss')
    make_plots(class_weight_history, f"{CASE} with class weight", 2, 'loss')
    plt.show()
    make_plots(no_class_weight_history, f"{CASE} no class weight", 0, 'accuracy')
    make_plots(no_class_weight_bias_history, f"{CASE} no class weight bias", 1, 'accuracy')
    make_plots(class_weight_history, f"{CASE} with class weight", 2, 'accuracy')
    plt.show()
    make_plots(no_class_weight_history, f"{CASE} no class weight", 0, 'precision')
    make_plots(no_class_weight_bias_history, f"{CASE} no class weight bias", 1, 'precision')
    make_plots(class_weight_history, f"{CASE} with class weight", 2, 'precision')
    plt.show()
    make_plots(no_class_weight_history, f"{CASE} no class weight", 0, 'recall')
    make_plots(no_class_weight_bias_history, f"{CASE} no class weight bias", 1, 'recall')
    make_plots(class_weight_history, f"{CASE} with class weight", 2, 'recall')
    plt.show()
    make_plots(no_class_weight_history, f"{CASE} no class weight", 0, 'auc')
    make_plots(no_class_weight_bias_history, f"{CASE} no class weight bias", 1, 'auc')
    make_plots(class_weight_history, f"{CASE} with class weight", 2, 'auc')
    plt.show()
    
if __name__ == '__main__':
    main()