#!/usr/bin/env python
import numpy as np
import paths as pt
from typing import List
from tools import file_reader, file_writer, preprocessor
from tools.classifiers import KnnClassifier, SvmClassifier, LrClassifier
from tools.classifiers import XgbClassifier, RfClassifier, MlpClassifier
import tensorflow as tf
from pathlib import Path
tf.get_logger().setLevel('ERROR')

NUM_ITER = 10
CASES = ["Complete", "Compliance", "Fall", "Fall_test"]

class Result:
  def __init__(self, name, result):
    self.name = name
    self.result = result

class CVResult:
  def __init__(self, acc, pre, rec, rocauc):
    self.acc = acc
    self.pre = pre
    self.rec = rec
    self.rocauc = rocauc
    
ATS_COLS = [str(i)+'Ats' for i in range(1, 10+1)] \
    + ['Cluster', 'LoanPeriod', 'NumberAts']
EX_COLS = [str(i)+'Ex' for i in range(1, 9+1)] + ['NumberEx']
CLF_NAMES = ["MLP", "LR", "XGB", "RF", "SVM", "KNN"]

def load_complete():
    ats = {str(i)+'Ats':str for i in range(1,11)}
    df = file_reader.read_csv(pt.PROCESSED_DATA_DIR,
                              'complete.csv',
                              converters=ats)
    return df

def load_fall():
    converters = {str(i)+'Ats':str for i in range(1,11)}
    df = file_reader.read_csv(pt.PROCESSED_DATA_DIR,
                              'fall.csv',
                              converters=converters)
    return df

def load_compliance():
    converters = {str(i)+'Ats':str for i in range(1,11)}
    df = file_reader.read_csv(pt.PROCESSED_DATA_DIR,
                              'compliance.csv',
                              converters=converters)
    return df

def load_fall_test():
    ex = {str(i)+'Ex':str for i in range(1, 9+1)}
    ats = {str(i)+'Ats':str for i in range(1, 10+1)}
    converters = {**ex, **ats}
    df = file_reader.read_csv(pt.PROCESSED_DATA_DIR,
                              'fall_test.csv',
                              converters=converters)
    return df

def main():
    for case in CASES:
        results_filename = f"{case} baseline results.txt"
        # Version 1
        with open(Path.joinpath(pt.REPORTS_DIR, results_filename), "w+") as text_file:
            text_file.write(f"{case} version 1 - without Ats and/or Ex columns")
            
        if case == "Complete":
            df = load_complete()
            X = df.drop(['Complete'], axis=1)   
            y = df['Complete']
            X = X.drop(ATS_COLS, axis=1)
        elif case == "Compliance":
            df = load_compliance()
            X = df.drop(['Compliance'], axis=1)   
            y = df['Compliance']
            X = X.drop(ATS_COLS, axis=1)
        elif case == "Fall":
            df = load_fall()
            X = df.drop(['Fall'], axis=1)
            y = df['Fall']
            X = X.drop(ATS_COLS, axis=1)
        else:
            df = load_fall_test()
            X = df.drop(['Fall'], axis=1)
            y = df['Fall']
            X = X.drop(ATS_COLS + EX_COLS, axis=1)
        
        X = np.array(X)
        y = np.array(y)
        
        y_pred_probas, results_mean, results_std = [], [], []
        for clf_name in CLF_NAMES:
            y_pred_proba, result_mean, result_std = train_clf(X, y, clf_name,
             results_filename, case, n_scale_cols=X.shape[1])
            y_pred_probas.append(y_pred_proba)
            results_mean.append(result_mean)
            results_std.append(result_std)
        make_plots(y, zip(CLF_NAMES, y_pred_probas, results_mean, results_std),
                   case, 1, "without Ats and/or Ex columns")
                
        # Version 2
        with open(Path.joinpath(pt.REPORTS_DIR, results_filename), "a") as text_file:
            text_file.write("\n\n")
            text_file.write(f"{case} version 2 - with embeddings")

        if case == "Complete":
            df = file_reader.read_csv(pt.PROCESSED_DATA_DIR, 'complete_emb.csv')
            X = df.drop(['Complete'], axis=1)
            y = df['Complete']
        elif case == "Compliance":
            df = file_reader.read_csv(pt.PROCESSED_DATA_DIR, 'compliance_emb.csv')
            X = df.drop(['Compliance'], axis=1)
            y = df['Compliance']
        elif case == "Fall":
            df = file_reader.read_csv(pt.PROCESSED_DATA_DIR, 'fall_emb.csv')
            X = df.drop(['Fall'], axis=1)
            y = df['Fall']
        else:
            df = file_reader.read_csv(pt.PROCESSED_DATA_DIR, 'fall_test_emb.csv')
            X = df.drop(['Fall'], axis=1)
            y = df['Fall']
        
        X = np.array(X)
        y = np.array(y)
        
        ats_cols = df.filter(regex='((\d+)[Ats])\w+', axis=1)
        ex_cols = df.filter(regex='((\d+)[Ex])\w+', axis=1)
        n_scale_cols = X.shape[1] - ats_cols.shape[1] - ex_cols.shape[1]
        
        y_pred_probas, results_mean, results_std = [], [], []
        for clf_name in CLF_NAMES:
            y_pred_proba, result_mean, result_std = train_clf(X, y, clf_name,
             results_filename, case, n_scale_cols=n_scale_cols)
            y_pred_probas.append(y_pred_proba)
            results_mean.append(result_mean)
            results_std.append(result_std)
        make_plots(y, zip(CLF_NAMES, y_pred_probas, results_mean, results_std),
                   case, 2, "with embeddings")
        
        # Version 3
        with open(Path.joinpath(pt.REPORTS_DIR, results_filename), "a") as text_file:
            text_file.write("\n\n")
            text_file.write(f"{case} version 3 - with counts")
            
        if case == "Complete":
            df = file_reader.read_csv(pt.PROCESSED_DATA_DIR, 'complete_count.csv')
            X = df.drop(['Complete'], axis=1)
            y = df['Complete']
        elif case == "Compliance":
            df = file_reader.read_csv(pt.PROCESSED_DATA_DIR, 'compliance_count.csv')
            X = df.drop(['Compliance'], axis=1)
            y = df['Compliance']
        elif case == "Fall":
            df = file_reader.read_csv(pt.PROCESSED_DATA_DIR, 'fall_count.csv')
            X = df.drop(['Fall'], axis=1)
            y = df['Fall']
        else:
            df = file_reader.read_csv(pt.PROCESSED_DATA_DIR, 'fall_test_count.csv')
            X = df.drop(['Fall'], axis=1)
            y = df['Fall']
        
        X = np.array(X)
        y = np.array(y)
        y_pred_probas, results_mean, results_std = [], [], []
        for clf_name in CLF_NAMES:
            y_pred_proba, result_mean, result_std = train_clf(X, y, clf_name,
             results_filename, case, n_scale_cols=X.shape[1])
            y_pred_probas.append(y_pred_proba)
            results_mean.append(result_mean)
            results_std.append(result_std)
        make_plots(y, zip(CLF_NAMES, y_pred_probas, results_mean, results_std),
                   case, 3, "with counts")

def train_clf(X: np.ndarray, y: np.ndarray, clf_name: str,
              results_filename: str, case: str, n_scale_cols:int=0) -> np.ndarray:
    y_pred_probas = 0
    y_pred_acc_mean, y_pred_pre_mean, y_pred_rec_mean, y_pred_rocauc_mean = [], [], [], []
    y_pred_acc_std, y_pred_pre_std, y_pred_rec_std, y_pred_rocauc_std = [], [], [], []
    for k in range(NUM_ITER):
        X = preprocessor.scale_data_standard(X, n_scale_cols)
        if clf_name == "MLP":
            results = MlpClassifier(X, y).evaluate(k)
        elif clf_name == "LR":
            results = LrClassifier(X, y).evaluate(k)
        elif clf_name == "XGB":
            results = XgbClassifier(X, y).evaluate(k)
        elif clf_name == "RF":
            results = RfClassifier(X, y).evaluate(k)
        elif clf_name == "SVM":
            results = SvmClassifier(X, y).evaluate(k)
        else:
            results = KnnClassifier(X, y).evaluate(k)
        make_and_print_scores(clf_name, k, results[0], results[1], results[2],
                              results[3], results[4], results_filename)
        y_pred_probas += results[5][:,1]
        y_pred_acc_mean.append(results[0].mean())
        y_pred_pre_mean.append(results[1].mean())
        y_pred_rec_mean.append(results[2].mean())
        y_pred_rocauc_mean.append(results[3].mean())
        y_pred_acc_std.append(results[0].std())
        y_pred_pre_std.append(results[1].std())
        y_pred_rec_std.append(results[2].std())
        y_pred_rocauc_std.append(results[3].std())
    y_pred_probas /= NUM_ITER
    results_mean = CVResult(y_pred_acc_mean, y_pred_pre_mean,
                            y_pred_rec_mean, y_pred_rocauc_mean)
    results_std = CVResult(y_pred_acc_std, y_pred_pre_std,
                           y_pred_pre_std, y_pred_rocauc_std)
    return y_pred_probas, results_mean, results_std
    
def make_and_print_scores(classifer_name: str, k: int, res_acc: List,
                          res_pre: List, res_rec: List, res_rocauc: List,
                          results_filename: str):
    results_mean = [Result("Accuracy", res_acc.mean()),
               Result("Precision", res_pre.mean()),
               Result("Recall", res_rec.mean()),
               Result("ROCAUC", res_rocauc.mean())]
    results_std = [Result("Accuracy", res_acc.std()),
               Result("Precision", res_pre.std()),
               Result("Recall", res_rec.std()),
               Result("ROCAUC", res_rocauc.std())]
    metrics = " - ".join(["{}: {:.4f} (+/- {:.2f})".format(r_m.name, r_m.result, r_s.result)
                          for r_m, r_s in (zip(results_mean, results_std) or [])])
    text ="\r{} K={}: {}".format(classifer_name, k, metrics)
    with open(Path.joinpath(pt.REPORTS_DIR, results_filename), "a") as text_file:
        text_file.write(text)

def make_plots(y_test: np.ndarray, results: np.ndarray,
               case_name: str, version_number: int, case_subtitle):
    roc_file_name = f"{case_name} version {version_number} - ROC curves.pdf"
    results_list = list(results)
    file_writer.write_roc_curve(y_test, results_list,
                               pt.REPORTS_PLOTS_DIR, roc_file_name, case_subtitle)
    file_writer.write_accuracy_plot(results_list, NUM_ITER, CLF_NAMES, pt.REPORTS_PLOTS_DIR,
                                 f"{case_name} version {version_number} - Accuracy.pdf", case_subtitle)
    file_writer.write_precision_plot(results_list, NUM_ITER, CLF_NAMES, pt.REPORTS_PLOTS_DIR,
                                 f"{case_name} version {version_number} - Precision.pdf", case_subtitle)
    file_writer.write_recall_plot(results_list, NUM_ITER, CLF_NAMES, pt.REPORTS_PLOTS_DIR, 
                                 f"{case_name} version {version_number} - Recall.pdf", case_subtitle)
    file_writer.write_rocauc_plot(results_list, NUM_ITER, CLF_NAMES, pt.REPORTS_PLOTS_DIR,
                                 f"{case_name} version {version_number} - ROCAUC.pdf", case_subtitle)
    file_writer.write_cm_plot(y_test, results_list[2][1], pt.REPORTS_PLOTS_DIR,
                              f"{case_name} version {version_number} - CM.pdf", "XGB - " f'{case_subtitle}')
    
if __name__ == '__main__':
    main()
