import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow.python.autograph.core.converter import Base
import xgboost as xgb
from sklearn.neighbors import KNeighborsClassifier
from sklearn.model_selection import cross_validate, cross_val_predict
from sklearn.model_selection import StratifiedKFold
from sklearn.ensemble import RandomForestClassifier
from sklearn.svm import SVC
from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVC
from tools import preprocessor
from abc import ABC, abstractmethod
from tensorflow.keras.wrappers.scikit_learn import KerasClassifier

class ClassifierResult():
    def __init__(self, acc, pre, rec, rocauc, prauc, probas):
        self.accuracy = acc
        self.precision = pre
        self.recall = rec
        self.rocauc = rocauc
        self.prauc = prauc
        self.probas = probas

class BaseClassifer(ABC):
    def __init__(self, X, y):
        """Initilizes inputs and targets variables"""
        self.X = X
        self.y = y
    
    @abstractmethod
    def make_model(self):
        """Makes a model"""
        
    def evaluate(self, k=0) -> ClassifierResult:
        skf = StratifiedKFold(n_splits=5, shuffle=True, random_state=k)
        metrics = ['accuracy', 'precision', 'recall', 'roc_auc', 'average_precision']
        model = self.make_model()
        res_validate = cross_validate(model, self.X, self.y, cv=skf, scoring=metrics)
        res_probas = cross_val_predict(model, self.X, self.y, cv=skf, method="predict_proba")
        result = ClassifierResult(res_validate['test_accuracy'], res_validate['test_precision'],
                                  res_validate['test_recall'], res_validate['test_roc_auc'],
                                  res_validate['test_average_precision'], res_probas)
        return result
    
class KnnClassifier(BaseClassifer):
    def make_model(self):
        return KNeighborsClassifier(n_neighbors=10,
                                    weights='distance',
                                    p=1)

class SvmClassifier(BaseClassifer):
    def make_model(self):
        return SVC(random_state=0,
                   class_weight="balanced",
                   probability=True)

class LrClassifier(BaseClassifer):
    def make_model(self):
        return LogisticRegression(max_iter=1000,
                                  class_weight="balanced",
                                  random_state=0)

class XgbClassifier(BaseClassifer):
    def make_model(self):
        neg, pos = np.bincount(self.y)
        scale_pos_weight = neg / pos
        params = {"n_estimators": 400,
                  "learning_rate": 0.1,
                  "max_depth": 4,
                  "scale_pos_weight": scale_pos_weight,
                  "objective": "binary:logistic",
                  "seed": 0,
                  "use_label_encoder": False,
                  "eval_metric": 'logloss'}
        return xgb.XGBClassifier(**params)

class RfClassifier(BaseClassifer):
    def make_model(self):
        return RandomForestClassifier(n_estimators=800,
                                      class_weight="balanced",
                                      random_state=0)
        
class MlpClassifier(BaseClassifer):
    def make_model(self):
        def make_keras_model():
            model = tf.keras.models.Sequential()
            model.add(tf.keras.layers.Dense(80,
                                            input_dim=self.X.shape[1],
                                            activation='relu'))
            model.add(tf.keras.layers.Dropout(0.35))
            model.add(tf.keras.layers.Dense(20, activation='relu'))
            model.add(tf.keras.layers.Dropout(0.15))
            model.add(tf.keras.layers.Dense(10, activation='relu'))
            model.add(tf.keras.layers.Dropout(0.15))
            model.add(tf.keras.layers.Dense(1, activation='sigmoid'))
            metrics = [
                tf.keras.metrics.BinaryAccuracy(name='accuracy'),
                tf.keras.metrics.Precision(name='precision'),
                tf.keras.metrics.Recall(name='recall'),
                tf.keras.metrics.AUC(name='roc_auc'),
                tf.keras.metrics.AUC(name='pr_auc', curve='PR'),
            ]
            model.compile(loss='binary_crossentropy',
                        optimizer="Adam",
                        metrics=metrics)
            return model
        
        neg, pos = np.bincount(self.y)
        class_weight = preprocessor.get_class_weight(neg, pos)
        return KerasClassifier(make_keras_model, epochs=20, batch_size=32,
                               class_weight=class_weight, verbose=False)