import os
from pathlib import Path
import pandas as pd
from typing import List
import matplotlib.pyplot as plt
import shap
import joblib
import numpy as np
import tensorflow as tf
import csv
import itertools
import pickle
import matplotlib
from sklearn.metrics import roc_curve, confusion_matrix

def write_csv(df: pd.DataFrame,
              path: Path,
              file_name: str,
              date_format: str='%d-%m-%Y',
              index: bool=False) -> None:
    df.to_csv(Path.joinpath(path, file_name),
              date_format=date_format,
              index=index)

def write_mapping(mapping: dict, path: Path, file_name: str) -> None:
    field_names = ['Ats', 'Embedding']
    with open(Path.joinpath(path, file_name), 'w', newline='', encoding='utf-8') as f:
        writer = csv.writer(f)
        writer.writerow(field_names)
        for key, value in mapping.items():
            writer.writerow([key, value])
    f.close()

def write_array(data, path, file_name) -> None:
    np.save(Path.joinpath(path, file_name), data)

def write_pickle(data: any, path: Path, file_name: str) -> None:
    with open(Path.joinpath(path, file_name), 'wb') as f:
        pickle.dump(data, f)

def write_joblib(model: any,
                path: Path,
                file_name: str) -> None:
    joblib.dump(model, Path.joinpath(path, file_name))

def write_tensorflow_model(model: tf.keras.Model,
                           path: Path) -> None:
    model.save(path)

def write_report(report, path, filename):
    report.to_file(Path.joinpath(path, filename))
    
def write_shap_summary_plot(shap_values: List[np.ndarray],
                            X_test: pd.DataFrame,
                            feature_names: List[str],
                            file_name: str,
                            path: Path,
                            plot_type: str) -> None:
    plt.close()
    shap.summary_plot(shap_values,
                      X_test,
                      feature_names=feature_names,
                      plot_type=plot_type,
                      show=False)
    
    plt.savefig(Path.joinpath(path, f'{file_name}.pdf'),
                dpi=300,
                bbox_inches = "tight")

def write_shap_importance_plot(features, importances, path, file_name):
    plt.close()
    plt.title(file_name)
    plt.xlabel('SHAP values')
    plt.ylabel('Feature')
    plt.barh(features[::-1], importances[::-1])
    plt.savefig(Path.joinpath(path, f'{file_name}.pdf'),
                dpi=300,
                bbox_inches = "tight")

def write_roc_curve(y_test: np.ndarray, results: List,
                    path: Path, file_name: str, case_subtitle: str):
    plt.close()
    plt.suptitle(os.path.splitext(file_name)[0])
    plt.title(case_subtitle)
    plt.ylabel("TPR")
    plt.xlabel("FPR")
    for result in results:
        fpr, tpr, _ = roc_curve(y_test, result[1])
        plt.plot(fpr, tpr, linewidth=2, label=result[0])
        plt.plot([0, 1], [0, 1], 'k--')
    plt.legend(loc="lower right")
    plt.savefig(Path.joinpath(path, file_name), dpi=300, bbox_inches = "tight")

def write_accuracy_plot(results: List, num_iter: int, clf_names: List[str],
                     path: Path, file_name: str, case_subtitle: str):
    plt.close()
    x = range(num_iter)
    fig, axs = plt.subplots(nrows=2, ncols=3, sharex=True, sharey=True)
    for i, ax in enumerate(axs.flat):
      ax.set_title(clf_names[i])
      ax.errorbar(x, results[i][2].acc, yerr=results[i][3].acc,
                  fmt='o', color='black', ecolor='lightgray',
                  elinewidth=3, capsize=0)
    fig.suptitle(f"{os.path.splitext(file_name)[0]} {case_subtitle}")
    plt.setp(axs[-1, :], xlabel='Seed')
    plt.setp(axs[:, 0], ylabel='Accuracy')
    plt.savefig(Path.joinpath(path, file_name), dpi=300, bbox_inches = "tight")

def write_precision_plot(results: List, num_iter: int, clf_names: List[str],
                     path: Path, file_name: str, case_subtitle: str):
    plt.close()
    x = range(num_iter)
    fig, axs = plt.subplots(nrows=2, ncols=3, sharex=True, sharey=True)
    for i, ax in enumerate(axs.flat):
      ax.set_title(clf_names[i])
      ax.errorbar(x, results[i][2].pre, yerr=results[i][3].pre,
                  fmt='o', color='black', ecolor='lightgray',
                  elinewidth=3, capsize=0)
    fig.suptitle(f"{os.path.splitext(file_name)[0]} {case_subtitle}")
    plt.setp(axs[-1, :], xlabel='Seed')
    plt.setp(axs[:, 0], ylabel='Precision')
    plt.savefig(Path.joinpath(path, file_name), dpi=300, bbox_inches = "tight")
    
def write_recall_plot(results: List, num_iter: int, clf_names: List[str],
                     path: Path, file_name: str, case_subtitle: str):
    plt.close()
    x = range(num_iter)
    fig, axs = plt.subplots(nrows=2, ncols=3, sharex=True, sharey=True)
    for i, ax in enumerate(axs.flat):
      ax.set_title(clf_names[i])
      ax.errorbar(x, results[i][2].rec, yerr=results[i][3].rec,
                  fmt='o', color='black', ecolor='lightgray',
                  elinewidth=3, capsize=0)
    fig.suptitle(f"{os.path.splitext(file_name)[0]} {case_subtitle}")
    plt.setp(axs[-1, :], xlabel='Seed')
    plt.setp(axs[:, 0], ylabel='Recall')
    plt.savefig(Path.joinpath(path, file_name), dpi=300, bbox_inches = "tight")
    
def write_rocauc_plot(results: List, num_iter: int, clf_names: List[str],
                     path: Path, file_name: str, case_subtitle: str):
    plt.close()
    x = range(num_iter)
    fig, axs = plt.subplots(nrows=2, ncols=3, sharex=True, sharey=True)
    for i, ax in enumerate(axs.flat):
      ax.set_title(clf_names[i])
      ax.errorbar(x, results[i][2].rocauc, yerr=results[i][3].rocauc,
                  fmt='o', color='black', ecolor='lightgray',
                  elinewidth=3, capsize=0)
    fig.suptitle(f"{os.path.splitext(file_name)[0]} {case_subtitle}")
    plt.setp(axs[-1, :], xlabel='Seed')
    plt.setp(axs[:, 0], ylabel='ROCAUC')
    plt.savefig(Path.joinpath(path, file_name), dpi=300, bbox_inches = "tight")
    
def write_cm_plot(y_test: np.ndarray, y_pred: np.ndarray,
                  path: Path, file_name: str, case_subtitle: str):
    plt.close()
    y_scores_new = (y_pred > 0.5)
    cm = confusion_matrix(y_test, y_scores_new)

    mask = None
    cm = (cm.T / np.sum(cm, axis=1)).T
    mask = np.empty_like(cm, dtype=bool)
    cm = (np.round(cm, decimals=2) * 100).astype(int)

    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        if cm[i, j] < 1:
            mask[i, j] = True
        else:
            mask[i, j] = False
            
    fig, ax = plt.subplots(figsize = (5,5))
    ax.set_xticks(range(2))
    ax.set_yticks(range(2))
    ax.set_xticklabels(range(2))
    ax.set_yticklabels(range(2))
    ax.tick_params(left=False, bottom=False)
    
    im = ax.imshow(cm, cmap="Greys")
 
    invert_threshold = True
    if mask is None:
        if invert_threshold == True:
            annotate_heatmap(im, valfmt="{x:.0f}", invertThreshold=True)
        else:
            annotate_heatmap(im, valfmt="{x:.0f}")
    else:
        if invert_threshold == True:
            annotate_heatmap(im, mask=mask, valfmt="{x:.0f}", invertThreshold=True)
        else:
            annotate_heatmap(im, mask=mask, valfmt="{x:.0f}")
            
    for _, spine in ax.spines.items():
        spine.set_visible(True)
    
    plt.ylabel("True label")
    plt.xlabel("Predicted label")
    plt.title(case_subtitle)
    plt.suptitle(os.path.splitext(file_name)[0])
    plt.savefig(Path.joinpath(path, file_name), dpi=300, bbox_inches = "tight")
    
def annotate_heatmap(im, data=None, valfmt="{x:.2f}",
					 textcolors=["black", "white"], textcoloridx=None,
					 threshold=None, invertThreshold=False, mask=None, **textkw):
	"""
	A function to annotate a heatmap.
	Parameters
	----------
	im
		The AxesImage to be labeled.
	data
		Data used to annotate.  If None, the image's data is used.  Optional.
	valfmt
		The format of the annotations inside the heatmap.  This should either
		use the string format method, e.g. "$ {x:.2f}", or be a
		`matplotlib.ticker.Formatter`.  Optional.
	textcolors
		A list or array of two color specifications.  The first is used for
		values below a threshold, the second for those above.  Optional.
	threshold
		Value in data units according to which the colors from textcolors are
		applied.  If None (the default) uses the middle of the colormap as
		separation.  Optional.
	**kwargs
		All other arguments are forwarded to each call to `text` used to create
		the text labels.
	"""

	if not isinstance(data, (list, np.ndarray)):
		data = im.get_array()

	# Normalize the threshold to the images color range.
	if threshold is not None:
		threshold = im.norm(threshold)
	else:
		threshold = im.norm(data.max())/2.

	# Set default alignment to center, but allow it to be
	# overwritten by textkw.
	kw = dict(horizontalalignment="center",
			  verticalalignment="center")
	kw.update(textkw)

	# Get the formatter in case a string is supplied
	if isinstance(valfmt, str):
		valfmt = matplotlib.ticker.StrMethodFormatter(valfmt)

	# Loop over the data and create a `Text` for each "pixel".
	# Change the text's color depending on the data.
	texts = []
	for i in range(data.shape[0]):
		for j in range(data.shape[1]):
			if textcoloridx is not None:
				if invertThreshold:
					kw.update(color=textcolors[int(im.norm(textcoloridx[i, j]) > threshold)])
				else:
					kw.update(color=textcolors[int(im.norm(textcoloridx[i, j]) < threshold)])
			else:
				if invertThreshold:
					kw.update(color=textcolors[int(im.norm(data[i, j]) > threshold)])
				else:
					kw.update(color=textcolors[int(im.norm(data[i, j]) < threshold)])
			if mask is not None:
				if not mask[i, j]:
					text = im.axes.text(j, i, valfmt(data[i, j], None), **kw)
			else:
				text = im.axes.text(j, i, valfmt(data[i, j], None), **kw)
				texts.append(text)

	return texts

def write_shap_explanation(expected_value: float,
                           shap_value: np.ndarray,
                           row: pd.Series,
                           feature_names: List[str],
                           file_name: str,
                           path: Path):
    shap.force_plot(expected_value, shap_value, row,
                    feature_names=feature_names,
                    show=False, matplotlib=True).savefig(Path.joinpath(path,
                                                            f'{file_name}.png'))